/*
 *  Citrusleaf
 *  cf_rb/citrusleaf.c - Ruby module
 *
 *  Copyright 2008-2009 by Citrusleaf.  All rights reserved.
 *  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE.  THE COPYRIGHT NOTICE
 *  ABOVE DOES NOT EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION.
 */
#include "citrusleaf/citrusleaf.h"
#include "ruby.h"

//
// Will marshall and return a newly allocated string type
//

static VALUE g_marshal_value = Qnil;
static ID g_marshal_dump_id = 0;
static ID g_marshal_load_id = 0;

VALUE
cCitrusleaf_marshal(VALUE v)
{
	// rb_funcall( VALUE ('this'), ID (huh), int nargs, .... args
	
// this evaulates to an object of type module (RUBY_T_MODULE)
	if (g_marshal_value == Qnil) {
		g_marshal_value = rb_eval_string("Marshal");
	}

	if (g_marshal_dump_id == 0) {	
		g_marshal_dump_id = rb_intern("dump");
	}
	
	return ( rb_funcall(g_marshal_value ,g_marshal_dump_id,1,v) );
}

VALUE
cCitrusleaf_unmarshal(VALUE v)
{
	if (g_marshal_value == Qnil) {
		g_marshal_value = rb_eval_string("Marshal");
	}
	if (g_marshal_load_id == 0) {	
		g_marshal_load_id = rb_intern("load");
	}
	
	return(rb_funcall(g_marshal_value , g_marshal_load_id , 1 , v));
}



/* _Citrusleaf_key_construct
 * Construct a key object */
static void
_Citrusleaf_key_construct(cl_object *obj_key, VALUE key)
{
	switch(TYPE(key)) {
		case T_STRING:
			citrusleaf_object_init_str2(obj_key, StringValuePtr(key), RSTRING_LEN(key));
			break;
		case T_FIXNUM:
		case T_BIGNUM:
			citrusleaf_object_init_int(obj_key, NUM2LONG(key));
			break;
		case T_NIL:
			rb_raise(rb_eRuntimeError, "nil not allowed as key");
			break;
		default: // blob up that which I don't understand
			{
			VALUE v = cCitrusleaf_marshal(key);
			citrusleaf_object_init_blob2(obj_key, StringValuePtr(v),RSTRING_LEN(v),CL_RUBY_BLOB);
			}
			break;
	}

	return;
}


//
// DRY! do the type mappings in one place
//
static int
cCitrusleaf_set_bin(cl_bin *bin, VALUE name, VALUE value)
{
	strcpy(bin->bin_name, StringValuePtr(name));
	switch (TYPE(value)) {
		case T_STRING:
			citrusleaf_object_init_str2(&bin->object, StringValuePtr(value), RSTRING_LEN(value));
			break;
		case T_FIXNUM:
		case T_BIGNUM:
			citrusleaf_object_init_int(&bin->object, NUM2LONG(value));
			break;
		case T_NIL:
			citrusleaf_object_init_null(&bin->object);
			break;
		case T_TRUE:
			citrusleaf_object_init_int(&bin->object, NUM2LONG(1));
			break;
		case T_FALSE:
			citrusleaf_object_init_int(&bin->object, NUM2LONG(0));
			break;
		default:
			{
				VALUE v = cCitrusleaf_marshal(value);
				citrusleaf_object_init_blob2(&bin->object, StringValuePtr(v),RSTRING_LEN(v),CL_RUBY_BLOB);
			}
			break;
	}
	return(0);
}

/* _Citrusleaf_hash2binv
 * An iterator to transform the contents of a Ruby hash into a bin vector */
static int
_Citrusleaf_hash2binv(VALUE key, VALUE value, VALUE extra)
{
	cl_bin *va = (cl_bin *)extra;

	// This is awful.  Sorry. Forwards to the next bin, I guess.
	while (0 != *(void **)va)
		va++;

	cCitrusleaf_set_bin(va, key, value);

	return( (int) ST_CONTINUE);
}


/* cCitrusleaf_init
 * Initializer */
static VALUE
cCitrusleaf_init(VALUE self)
{
	rb_iv_set(self, "@timeout", INT2FIX(500));
	rb_iv_set(self, "@following", Qtrue);
	rb_iv_set(self, "@writepolicy", INT2FIX(CL_WRITE_RETRY));

	return(self);
}


/* cCitrusleaf_new
 * A constructor for a new Citrusleaf */
static VALUE
cCitrusleaf_new(VALUE class, VALUE self)
{
	cl_cluster *cptr = citrusleaf_cluster_create();
	VALUE c = Data_Wrap_Struct(class, 0, citrusleaf_cluster_destroy, cptr);
	rb_obj_call_init(c, 0, 0);

	return(c);
}


/* cCitrusleaf_set_following
 * Allow (or disallow) a cluster from following node membership changes */
static VALUE
cCitrusleaf_set_following(VALUE self, VALUE arg)
{
	cl_cluster *c;
	Data_Get_Struct(self, cl_cluster, c);

	if (arg == Qtrue) {
		rb_iv_set(self, "@following", Qtrue);
		citrusleaf_cluster_follow(c, true);
	} else {
		rb_iv_set(self, "@following", Qfalse);
		citrusleaf_cluster_follow(c, false);
	}

	return(arg);
}


/* cCitrusleaf_get_following
 * Get the following status */
static VALUE
cCitrusleaf_get_following(VALUE self)
{
	return(rb_iv_get(self, "@following"));
}


/* cCitrusleaf_set_timeout
 * Set the per-operation timeout, in msec */
static VALUE
cCitrusleaf_set_timeout(VALUE self, VALUE timeout)
{
	Check_Type(timeout, T_FIXNUM);

	rb_iv_set(self, "@timeout", timeout);

	return(timeout);
}


/* cCitrusleaf_get_timeout
 * Get the per-operation timeout value */
static VALUE
cCitrusleaf_get_timeout(VALUE self)
{
	return(rb_iv_get(self, "@timeout"));
}


/* cCitrusleaf_set_writepolicy
 * Set the write policy */
static VALUE
cCitrusleaf_set_writepolicy(VALUE self, VALUE policy)
{
	Check_Type(policy, T_FIXNUM);

	rb_iv_set(self, "@writepolicy", policy);

	return(policy);
}


/* cCitrusleaf_get_writepolicy
 * Get the write policy */
static VALUE
cCitrusleaf_get_writepolicy(VALUE self)
{
	return(rb_iv_get(self, "@writepolicy"));
}


/* cCitrusleaf_add_node
 * Add a node to a Citrusleaf cluster */
static VALUE
cCitrusleaf_add_node(VALUE self, VALUE host, VALUE port)
{
	cl_cluster *c;
	Check_Type(host, T_STRING);
	Check_Type(port, T_FIXNUM);
	Data_Get_Struct(self, cl_cluster, c);

	citrusleaf_cluster_add_host(c, StringValuePtr(host), FIX2INT(port), 0);

	return(Qnil);
}


/* cCitrusleaf_get
 * Retrieve an object from a Citrusleaf cluster 
   arg1 - string, namespace
   arg2 - string, bin
   arg3 - object, key
   arg4 - optional - bin. either array of bin names, or string bin name, or nothing - means get all
*/
 static VALUE
cCitrusleaf_get(int argc, VALUE *args, VALUE self)
{
	cl_cluster *c;
	cl_object obj_key;
	int rv;
	cl_bin *binv;
	int binv_sz = 0;
	int i = 0;
	uint32_t generation = 0;
	
	bool binv_free = false;
	bool hash_return = true;
	
	if (argc < 3) 
		rb_raise(rb_eRuntimeError, "get must have at least 3 arguments");
	if (argc > 4)
		rb_raise(rb_eRuntimeError, "put has too many arguments: 4 max");
	
	VALUE namespace = args[0];
	VALUE set = args[1];
	_Citrusleaf_key_construct(&obj_key, args[2]);
    VALUE binz = argc >= 4 ? args[3] : T_NIL;
	
	Check_Type(namespace, T_STRING);
	Check_Type(set, T_STRING);
	
	Data_Get_Struct(self, cl_cluster, c);

	// backward compatibility...
	if ((TYPE(binz) == T_ARRAY) && (RARRAY_LEN(binz) == 0)) binz = Qnil;
	
	if (TYPE(binz) == T_ARRAY) {
		binv_sz = RARRAY_LEN(binz);
		binv_free = false;
	}
	else if (TYPE(binz)==T_STRING) {
		binv_sz = 1;
		binv_free = false;
	}
	cl_bin binav[binv_sz];
	binv = binav;	
	if (binv_sz) {
		for (i = 0; i< binv_sz; i++) {
			citrusleaf_object_init(&binav[i].object);
		}
	}

	if (binz == T_NIL) {
		/* We got passed an empty array; get all bins - library mallocs binv */
		rv = citrusleaf_get_all(c, StringValuePtr(namespace), StringValuePtr(set), &obj_key, &binv, &binv_sz, rb_iv_get(self, "@timeout"), &generation);
		binv_free = true;
	}
	else if (TYPE(binz) == T_ARRAY) {

		/* Set up the bin vector using the bin array we were provided */
		VALUE *ap = RARRAY_PTR(binz);
		for (i = 0; i < binv_sz; i++) {
			Check_Type(ap[i], T_STRING);
			strcpy(binav[i].bin_name, StringValuePtr(ap[i]));
		}

		rv = citrusleaf_get(c, StringValuePtr(namespace), StringValuePtr(set), &obj_key, binav, binv_sz, rb_iv_get(self, "@timeout"), &generation);

	}
	else if (TYPE(binz == T_STRING)) {
		
		/* Set up the bin vector using the bin array we were provided */
		strcpy(binav[0].bin_name, StringValuePtr(binz));

		rv = citrusleaf_get(c, StringValuePtr(namespace), StringValuePtr(set), &obj_key, binav, 1, rb_iv_get(self, "@timeout"), &generation);
		hash_return = false;
	}
	else {	
		rb_raise(rb_eRuntimeError, "get 4th param must be nil, array of strings, or string");
	}

	// return value
	VALUE rv_a = rb_ary_new2(3);
	rb_ary_store(rv_a, 0, INT2NUM(rv));
	rb_ary_store(rv_a, 1, INT2NUM(0));

	
	/* Construct a hash of results and return it, or return an error code */
	VALUE r = Qnil;
	if (CITRUSLEAF_OK == rv) {
		if (hash_return) {
			r = rb_hash_new();
			int i;
			for (i = 0; i < binv_sz; i++) {
				switch(binv[i].object.type) {
					case CL_NULL:
						rb_hash_aset(r, rb_str_new2(binv[i].bin_name), Qnil);
						break;
					// there's no real difference between blobs and strings in ruby
					case CL_STR:
					case CL_BLOB:
						rb_hash_aset(r, rb_str_new2(binv[i].bin_name), rb_str_new(binv[i].object.u.str, binv[i].object.sz));
						break;
					case CL_INT:
						rb_hash_aset(r, rb_str_new2(binv[i].bin_name), INT2NUM(binv[i].object.u.i64));
						break;
					case CL_RUBY_BLOB:
						{
						// the pain of this, creating a new string just to do the marshalling ---
						// really just want a temporary string buffer to point into the right place...
						VALUE s_v = rb_str_new(binv[i].object.u.str, binv[i].object.sz);
						VALUE m_v = cCitrusleaf_unmarshal(s_v);
						rb_hash_aset(r, rb_str_new2(binv[i].bin_name), m_v);
						}
						break;

					case CL_JAVA_BLOB:
					case CL_PYTHON_BLOB:
					case CL_CSHARP_BLOB:
						/* FIXME blobs --- should warn or specificy somehow what kind of blobs are involved */
					default:
						rb_hash_aset(r, rb_str_new2(binv[i].bin_name), Qnil);
						rb_raise(rb_eRuntimeError, "unknown object type %d returned",binv[i].object.type);
						break;
				}
			}
		}
		else {
			// DRY FAIL
			switch(binv[0].object.type) {
				case CL_STR:
					r = rb_str_new(binv[0].object.u.str, binv[0].object.sz);
					break;
				case CL_INT:
					r = INT2NUM(binv[0].object.u.i64);
					break;
				case CL_RUBY_BLOB:
					{
					// the pain of this, creating a new string just to do the marshalling ---
					// really just want a temporary string buffer to point into the right place...
					VALUE s_v = rb_str_new(binv[0].object.u.str, binv[0].object.sz);
					r = cCitrusleaf_unmarshal(s_v);
					}
					break;
				default:
					rb_raise(rb_eRuntimeError, "unknown object type %d returned from server",binv[0].object.type);
					break;
			}
		}
		
		/* The binv is no longer needed, so we can safely destroy it */
		if (true == binv_free)
			free(binv);
	}
	rb_ary_store(rv_a, 2, r);
	
	return(rv_a);
	
}

		
static VALUE
cCitrusleaf_put_hash(cl_cluster *c, char *ns, char *set, cl_object *obj_key, VALUE valueh )
{
	int binv_sz = NUM2INT(rb_funcall(valueh, rb_intern("size"), 0, NULL));
	
	cl_bin binv[binv_sz];
	memset(binv, 0, sizeof(binv));
	rb_hash_foreach(valueh, _Citrusleaf_hash2binv, (VALUE)&binv);
	int rv = citrusleaf_put(c, ns, set, obj_key, binv, binv_sz, 0); // todo: put back in write policy ,etc

	return(INT2NUM(rv));
}


// Array could be an array of tuples,
// or an array of "name" "value" "name" "value" ?

static VALUE
cCitrusleaf_put_array(cl_cluster *c, char *ns, char *set, cl_object *obj_key, VALUE valuea )
{
	int rv;
	int bina_sz = RARRAY_LEN(valuea);
	if (bina_sz == 0)
		rb_raise(rb_eRuntimeError, "setting null size array illegal");
	VALUE *ap = RARRAY_PTR(valuea);
	
	// array of tuples
	if ( TYPE(ap[0]) == T_ARRAY) {
		cl_bin bina[bina_sz];
		int i;
		for (i=0;i<bina_sz;i++) {
			if (TYPE(ap[i]) != T_ARRAY)
				rb_raise(rb_eRuntimeError, "setting with array, but not array of tuples");
			if (RARRAY_LEN(ap[i]) != 2)
				rb_raise(rb_eRuntimeError, "setting with array, but element not a 2-ple");
			VALUE *elta = RARRAY_PTR(ap[i]);
			cCitrusleaf_set_bin(&bina[i], elta[0], elta[1]);
		}
		
		rv = citrusleaf_put(c, ns, set, obj_key, bina, bina_sz, 0); // todo: put back in write policy ,etc
		
	}
	else {
		if (bina_sz & 1 != 0)
			rb_raise(rb_eRuntimeError, "setting with array of bins and values, but odd number, illegal");
		bina_sz = bina_sz >> 1;
		cl_bin bina[bina_sz];
		int i;
		for (i=0;i<bina_sz;i++) {
			VALUE name = ap[ i * 2 ];
			VALUE value = ap[ (i * 2)+1 ];
			cCitrusleaf_set_bin(&bina[i], name, value);
		}			
		
		rv = citrusleaf_put(c, ns, set, obj_key, bina, bina_sz, 0); // todo: put back in write policy ,etc
	}

	return(INT2NUM(rv));	
}

static VALUE
cCitrusleaf_put_pair(cl_cluster *c, char *ns, char *set, cl_object *obj_key, VALUE bin, VALUE value)
{
	cl_bin clbin;
	strcpy(clbin.bin_name, StringValuePtr(bin));
	cCitrusleaf_set_bin(&clbin, bin, value);
	
	int rv = citrusleaf_put(c, ns, set, obj_key, &clbin, 1, 0); // todo: put back in write policy ,etc
	
	return(INT2NUM(rv));
}



/* cCitrusleaf_put
 * Set an object in a Citrusleaf cluster */
static VALUE
cCitrusleaf_put(int argc, VALUE *args, VALUE self)
{
	cl_cluster *c;
	cl_object obj_key;
	VALUE rv;
	
	if (argc < 4) 
		rb_raise(rb_eRuntimeError, "put must have at least 4 arguments");
	if (argc > 5)
		rb_raise(rb_eRuntimeError, "put has too many arguments: 5 max");
	
	VALUE namespace = args[0];
	VALUE set = args[1];
	_Citrusleaf_key_construct(&obj_key, args[2]);

	Check_Type(namespace, T_STRING);
	Check_Type(set, T_STRING);
	
	Data_Get_Struct(self, cl_cluster, c);
	
	if ( argc == 4 ) {
		if (TYPE(args[3])==T_HASH) {
			rv = cCitrusleaf_put_hash(c, StringValuePtr(namespace), StringValuePtr(set), &obj_key, args[3]);
		}
		else if (TYPE(args[3])==T_ARRAY) {
			rv = cCitrusleaf_put_array(c, StringValuePtr(namespace), StringValuePtr(set), &obj_key, args[3]);			
		}
		else
			rb_raise(rb_eRuntimeError,"put with 4 arguments should be array or hash");
	}
	if (argc == 5) {
		Check_Type(args[3],T_STRING);
		rv = cCitrusleaf_put_pair(c, StringValuePtr(namespace), StringValuePtr(set), &obj_key, args[3], args[4]);			
	}
	return(rv);
}



/* cCitrusleaf_delete
 * Delete an object from a Citrusleaf cluster */
static VALUE
cCitrusleaf_delete(VALUE self, VALUE namespace, VALUE set, VALUE key)
{
	cl_cluster *c;
	cl_object obj_key;
	int rv;
	Check_Type(namespace, T_STRING);
	Check_Type(set, T_STRING);
	Data_Get_Struct(self, cl_cluster, c);

	_Citrusleaf_key_construct(&obj_key, key);
	rv = citrusleaf_delete(c, StringValuePtr(namespace), StringValuePtr(set), &obj_key, 0);

	return(Qnil);
}


/* Init_citrusleaf
 * Called when the module is loaded */
VALUE mCitrusleaf = Qnil;
void
Init_Citrusleaf()
{
	mCitrusleaf = rb_define_module("CitrusleafConnector");

	if (0 != citrusleaf_init())
		rb_raise(rb_eRuntimeError, "unable to initialize module");

	VALUE c;
	c = rb_define_class("Citrusleaf", rb_cObject);
	rb_define_singleton_method(c, "new", cCitrusleaf_new, 0);
	rb_define_const(c, "WritePolicy_Async", INT2FIX(CL_WRITE_ASYNC));
	rb_define_const(c, "WritePolicy_Oneshot", INT2FIX(CL_WRITE_ONESHOT));
	rb_define_const(c, "WritePolicy_Retry", INT2FIX(CL_WRITE_RETRY));
	rb_define_const(c, "WritePolicy_Assured", INT2FIX(CL_WRITE_ASSURED));
	rb_define_method(c, "initialize", cCitrusleaf_init, 0);
	rb_define_method(c, "add_node", cCitrusleaf_add_node, 2);
	rb_define_method(c, "set_following", cCitrusleaf_set_following, 1);
	rb_define_method(c, "get_following", cCitrusleaf_get_following, 0);
	rb_define_method(c, "set_timeout", cCitrusleaf_set_timeout, 1);
	rb_define_method(c, "get_timeout", cCitrusleaf_get_timeout, 0);
	rb_define_method(c, "set_writepolicy", cCitrusleaf_set_writepolicy, 1);
	rb_define_method(c, "get_writepolicy", cCitrusleaf_get_writepolicy, 0);
	rb_define_method(c, "get", cCitrusleaf_get, -1);
	rb_define_method(c, "put", cCitrusleaf_put, -1);
	rb_define_method(c, "delete", cCitrusleaf_delete, 3);

	return;
}
